//! Handle information relating to NZZ issues.

use std::sync::Arc;

use anyhow::Result;
use serde::{Deserialize, Serialize};
use time::Date;
use tokio::{spawn, sync::Semaphore, task::JoinSet};
use tracing::info;

const SEARCH_URL: &str = "https://zeitungsarchiv.nzz.ch/solr-epaper-search/1.0/search";
const ISSUE_URL: &str = "https://zeitungsarchiv.nzz.ch/archive/1.0/getPages";
const MAX_DOWNLOADS: usize = 4;

#[derive(Debug, Serialize, Deserialize)]
struct SearchData {
    query: String,
    offset: u32,
    #[serde(rename = "sortField")]
    sort_field: String,
    #[serde(rename = "sortOrder")]
    sort_order: String,
    #[serde(
        rename = "startDate",
        serialize_with = "crate::date::serialize",
        deserialize_with = "crate::date::deserialize"
    )]
    start_date: Date,
    #[serde(
        rename = "endDate",
        serialize_with = "crate::date::serialize",
        deserialize_with = "crate::date::deserialize"
    )]
    end_date: Date,
}

#[derive(Debug, Serialize, Deserialize)]
struct SearchResult {
    data: SearchInfo,
}

#[derive(Debug, Serialize, Deserialize)]
struct SearchInfo {
    total: u32,
    offset: u32,
    #[serde(rename = "pageSize")]
    page_size: u32,
    #[serde(rename = "resData")]
    res_data: Vec<IssueData>,
}

#[derive(Debug, Serialize, Deserialize)]
struct IssueData {
    #[serde(rename = "editionId")]
    edition_id: u32,
    #[serde(rename = "pageNumber")]
    page_nr: u32,
    #[serde(
        rename = "publicationDate",
        deserialize_with = "crate::date::deserialize"
    )]
    publication_date: Date,
}

#[derive(Debug, Serialize, Deserialize)]
struct PagesResult {
    data: PagesInfo,
}

#[derive(Debug, Serialize, Deserialize)]
struct PagesInfo {
    pages: Vec<Page>,
}

#[derive(Debug, Serialize, Deserialize)]
struct Page {
    #[serde(rename = "pmPageNumber")]
    page_nr: u32,
    #[serde(rename = "pageDocUrl")]
    doc: PageDoc,
}

#[derive(Debug, Serialize, Deserialize)]
struct PageDoc {
    #[serde(rename = "HIGHRES")]
    link: PageHighRes,
}

#[derive(Debug, Serialize, Deserialize)]
struct PageHighRes {
    url: String,
}

/// A single NZZ issue.
#[derive(Debug, Clone)]
pub struct Issue {
    /// Date of publication.
    pub publication_date: Date,
    /// ordered vector of page urls in the issue.
    pub pages: Vec<String>,
}

impl SearchData {
    pub fn new(offset: u32, start_date: Date, end_date: Date) -> Self {
        Self {
            query: "".to_string(),
            offset,
            sort_field: "media_ts".to_string(),
            sort_order: "desc".to_string(),
            start_date,
            end_date,
        }
    }
}

/// Search all issues between `from` and `to` (inclusive) using an `offset` into the results.
async fn offset_search(offset: u32, cookie: &str, from: Date, to: Date) -> Result<SearchInfo> {
    let data = SearchData::new(offset, from, to);
    let result: SearchResult = reqwest::Client::new()
        .post(SEARCH_URL)
        .header("Cookie", cookie)
        .json(&data)
        .send()
        .await?
        .json()
        .await?;
    Ok(result.data)
}

/// Only keep first pages, they are enough to get the edition id.
fn filter_issues(unfiltered_issues: Vec<IssueData>) -> Vec<IssueData> {
    unfiltered_issues
        .into_iter()
        .filter(|info| info.page_nr == 1)
        .collect()
}

/// Search all issues between `from` and `to` (inclusive) respecting pagination.
async fn search(cookie: &str, from: Date, to: Date) -> Result<Vec<IssueData>> {
    info!("looking for issues between {from} to {to}");
    let mut result = offset_search(0, cookie, from, to).await?;
    let mut issues: Vec<IssueData> = filter_issues(result.res_data);

    while result.offset + result.page_size < result.total {
        result = offset_search(result.offset + result.page_size, cookie, from, to).await?;
        issues.extend(filter_issues(result.res_data));
    }

    Ok(issues)
}

/// Fetch all page urls for the issue with eddition id `edition_id` and order them by page number.
async fn build_pages(cookie: &str, edition_id: u32) -> Result<Vec<String>> {
    let result: PagesResult = reqwest::Client::new()
        .post(ISSUE_URL)
        .header("Cookie", cookie)
        .json(&serde_json::json!({
            "editionId": edition_id,
        }))
        .send()
        .await?
        .json()
        .await?;

    let mut pages: Vec<(u32, String)> = result
        .data
        .pages
        .into_iter()
        .map(|page| (page.page_nr, page.doc.link.url))
        .collect();
    pages.sort_by(|a, b| a.0.cmp(&b.0));
    let pages = pages.into_iter().map(|page| page.1).collect();

    Ok(pages)
}

/// Fetch all page urls for `issues`.
async fn build_issues(cookie: &str, issues: Vec<IssueData>) -> Result<Vec<Issue>> {
    let mut hydrated_issues = Vec::new();
    let permits = Arc::new(Semaphore::new(MAX_DOWNLOADS));

    let mut jobs = JoinSet::new();
    for issue in issues {
        let permits = permits.clone();
        let cookie = cookie.to_string();

        let job: tokio::task::JoinHandle<Result<Issue>> = spawn(async move {
            let _permit = permits.acquire().await.unwrap();

            info!(
                "fetching page information for issue {}",
                issue.publication_date
            );
            let pages = build_pages(&cookie, issue.edition_id).await?;
            Ok(Issue {
                publication_date: issue.publication_date,
                pages,
            })
        });
        jobs.spawn(job);
    }

    while let Some(res) = jobs.join_next().await {
        let issue: Issue = res???;
        hydrated_issues.push(issue);
    }

    Ok(hydrated_issues)
}

/// Fetch issue information in the date range `from`- `to` (inclusive) using `cookie`for
/// authentication.
pub async fn fetch(cookie: &str, from: Date, to: Date) -> Result<Vec<Issue>> {
    let issues = search(cookie, from, to).await?;
    let issues = build_issues(cookie, issues).await?;

    Ok(issues)
}
